/*
  This file is part of AyyiGtk. http://ayyi.org
  copyright (C) 2004-2013 Tim Orford <tim@orford.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 3.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*

parts
-----

TODO proper support for midi parts
	-access to the shm Region is now more difficult. Solutions:
		1- store a pointer to the Region in gpart. <-- current favourite.
		2- store a pointer to get_region in gpart.
	-split gpart struct into 3: base, audio, midi

*/
#define __am_private__
#include <gtk/gtk.h>
#include <model/common.h>
#include <ayyi/ayyi_time.h>
#include <ayyi/ayyi_typedefs.h>
#include <ayyi/ayyi_client.h>
#include <ayyi/ayyi_song.h>
#include <ayyi/ayyi_action.h>
#include <ayyi/ayyi_msg.h>
#include <ayyi/ayyi_dbus_proxy.h>
#include <ayyi/ayyi_shm.h>

#include <model/model_types.h>
#include <model/am_message.h>
#include <model/am_track.h>
#include <model/am_song.h>
#include <model/am_pool.h>
#include <model/am_midi_part.h>
#include <model/am_part.h>
#include <model/am_time.h>

extern int            idle_ref;
extern int            debug;
extern Observer       observer;

extern gboolean       is_song_shm_local        (void*);

static void           part_label_sync          (AMPart*);
static gboolean      _am_part_sync_name        (AMPart*);
static gboolean       am_part_check_bounds     (AMPart*);
static AMTrack*      _am_part_get_ayyi_track   (AMPart*);
static bool          _am_part_sync_pos         (AMPart*);
static PoolItem*      am_part__get_pool_item   (AMPart*);
static uint32_t       am_part__get_auto_fg_colour (AMPart*);


#include "am_midi_part.c"


AMPart*
am_part_new()
{
	AMPart* part = g_new0(AMPart, 1);

	part->ident.type = AYYI_OBJECT_AUDIO_PART;
	part->ident.idx = -1;
	part->fg_colour = am_part__get_auto_fg_colour(part);

	pos_set(&part->length, 8, 0, 0);

	return part;
}


void
am_part__unref(AMPart* part)
{
	void am_part__free(AMPart* part)
	{
		PF2;
		g_free(part);
	}

	part->ref_count--;
	if(part->ref_count < 1 && part->ayyi->id && model_collection_find_by_id(am_parts, part->ayyi->id)) gwarn("freeing a part that is in the partlist!");
	if(part->ref_count < 1) am_part__free(part);
}


typedef struct
{
	AMPart*      part;
	AMTrack*     track;
	AyyiHandler  user_callback;
	gpointer     user_data;

} PartCallbackData;
#define callback_part ((PartCallbackData*)action->app_data)->part

static AyyiAction*
part_action_new(AMPart* part, AyyiHandler callback, gpointer user_data, char* format, ...)
{
	char name[AYYI_NAME_MAX];
	va_list argp;
	va_start(argp, format);
	vsnprintf(name, AYYI_NAME_MAX - 1, format, argp);
	va_end(argp);

	AyyiAction* a = ayyi_action_new(name);

	PartCallbackData* data = a->app_data = g_new0(PartCallbackData, 1);
	data->part = part;
	data->user_callback = callback;
	data->user_data = user_data;

	part->ref_count++;

	return a;
}


void
am_part__move(AMPart* part, GPos* pos, AMTrack* new_track, AyyiHandler user_callback, gpointer user_data)
{
	//note: the track is set in the callback as a separate operation. If new_track is NULL, the track is not changed.

	if (!part)               { gerr ("bad part arg: %p.", part); return; }
	if (pos->beat > 1000)    { gerr ("bad song position.");      return; }
	if (!gpos_is_valid(pos)) { gerr ("invalid song position.");  return; }
	if (!((AyyiSongService*)ayyi.service)->song) return;

	void
	am_part_move_done(AyyiAction* action)
	{
		//as well as tidying up following the transaction, this fn also sets the new Part track number.

		PartCallbackData* d = action->app_data;
		AMPart* part = callback_part;
		g_return_if_fail(part);

		dbg(2, "part=%s", part->name);
		_am_part_sync_pos(part);
		am_part_check_bounds(part);

		char buff[AYYI_BBST_MAX];
		pos2bbst(&part->start, buff);
		log_print(0, "Part moved to %s.", buff);

		if(d->track){ // track has changed
			//dbg (0, "track=%i->%i", old_t->track_num, t);
			am_part__set_track(part, d->track, d->user_callback, d->user_data);

		}else{
			g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_POS_X, NULL);
			call(d->user_callback, (AyyiIdent){action->obj.type, action->ret.idx}, &action->ret.error, d->user_data);
		}

#warning am_part__unref causing segfault
//		am_part__unref(part);
	}

	if(new_track){
		dbg(0, "new track: %s", new_track->name);
		part->track = new_track; // ensure that model data reflects the updated track before the operation is complete.
		//TODO if request fails, restore the old track.
	}

	AyyiAction* a        = part_action_new(part, user_callback, user_data, "part move %u", part->ayyi->shm_idx);
	a->callback          = am_part_move_done;
	PartCallbackData* d  = a->app_data;
	d->track             = new_track;

	char buff[AYYI_BBST_MAX]; pos2bbst(pos, buff);
	dbg (0, "part_idx=%i pos=%s", part->ayyi->shm_idx, buff);

	/*

	currently we are not using the ardour id as an identifier, as it is 64 bit....
	//TODO the ayyi id is now 64 bit so we CAN use it...?

	unlike other functions, we are using the sample number!
	what time format should we use? we should maybe standardise on the higher resolution song_pos?
	Or are we going to allow either/both?

	*/

	int sample = pos2samples(pos);

	a->op           = AYYI_SET;
	a->obj.type     = PART_IS_AUDIO(part) ? AYYI_OBJECT_AUDIO_PART : AYYI_OBJECT_MIDI_PART;
	a->obj_idx.idx1 = part->ayyi->shm_idx;
	a->prop         = AYYI_STIME;
	a->i_val        = sample;

	ayyi_action_execute(a);
}


void
am_part__set_track(AMPart* part, AMTrack* new_track, AyyiHandler user_callback, gpointer user_data)
{
	g_return_if_fail(part);
	g_return_if_fail(new_track);
	PF;

	void
	am_part__set_track_done(AyyiAction* action)
	{
		g_return_if_fail(action);
		PF;

		PartCallbackData* d = action->app_data;
		AyyiIdent id = {action->obj.type, action->ret.idx};

		AMTrack* new_track = ((PartCallbackData*)action->app_data)->track;
		AMPart* part = callback_part;
		g_return_if_fail(part);
		part->track = new_track;
		dbg(2, "track=%i", part->track->track_num);
																									dbg(0, "track=%i", part->track->track_num);

		if(am_part__sync(part, AYYI_NO_PROP)) g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_POS_X, NULL);

		if(d) call(d->user_callback, id, &action->ret.error, d->user_data);

		am_part__unref(part);
	}

	AyyiAction* a = part_action_new(part, user_callback, user_data, "part set track (part %u --> %s)", part->ayyi->shm_idx, new_track->name);
	a->ret.idx    = part->ayyi->shm_idx; //TODO should not be necc here. Need to clarify what ret.idx should contain when no _new_ object returned.
	a->callback   = am_part__set_track_done;
	((PartCallbackData*)a->app_data)->track = new_track;
	dbg(0, "part_idx=%i t=%i", part->ayyi->shm_idx, new_track->track_num);

	ayyi_set_obj(a, PART_IS_AUDIO(part) ? AYYI_OBJECT_AUDIO_PART : AYYI_OBJECT_MIDI_PART, part->ayyi->shm_idx, AYYI_TRACK, new_track->ident.idx);
}


void
am_part__trim_left(AMPart* part, AyyiHandler callback, gpointer user_data)
{
	g_return_if_fail(part);

	void
	part_trim_left_done(AyyiAction* action)
	{
		PF;
		AMPart* part = callback_part;
		g_return_if_fail(part);

		if(am_part__sync(part, AYYI_NO_PROP)) g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_POS_X & CHANGE_INSET, NULL);

		am_part__unref(part);
	}

	uint64_t mu = pos2mu(&part->start);
	uint32_t region_idx = part->ayyi->shm_idx;

	AyyiAction* a = part_action_new(part, callback, user_data, "region trim left. id %u", region_idx);
	a->callback   = part_trim_left_done;

	char buff[32]; mu2bbst(mu, buff);
	dbg (0, "new_left=%Lumu (%Li samples) %s", mu, ayyi_mu2samples(mu), buff);
	ayyi_set_obj(a, AYYI_OBJECT_PART, region_idx, AYYI_INSET, mu);
}


void
am_part__split(AMPart* part, AyyiSongPos* _pos, AyyiHandler user_callback, gpointer user_data)
{
	//After the request is initiated, the original part is truncated, and a new part is created.
	//-the main callback is a reply to our request.
	//-the second is an instance handler for the original part. It is called before the main callback. It doesnt appear to be very useful.

	void
	am_part__split_done(AyyiAction* action)
	{
		PF;
		PartCallbackData* d = action->app_data;
		AyyiIdent id = {action->obj.type, action->ret.idx};

		if(d) call(d->user_callback, id, &action->ret.error, d->user_data);

		if(action->ret.error){
			//let upstream display the error
			/*
			dbg(0, "error=%p", action->ret.error);
			warn_gerror(__func__, &action->ret.error);
			*/
		}else{

			log_print(LOG_OK, "part split");
		}

		am_parts__emit ("selection-change", NULL);

		AMPart* part = callback_part;
		g_return_if_fail(part);
#warning looks like unreffing wrong address
dbg(0, "part=%p ref_count=%i", part, part->ref_count);
//		am_part__unref(part);
	}

#ifdef XAAA
	struct _data
	{
		AyyiHandler callback;
		AyyiHandler handler;
	};

	void
	am_part__split_handler(AyyiIdent obj, GError** error, gpointer user_data)
	{
		//handler for AYYI_SET on our object. Unfortunately it is not suitable to call the user_callback as it is called before generic handlers have run.

		struct _data* data = user_data;
		if(data){
			//dbg(0, "handler=%p callback=%p", data->handler, data->callback);

			call(data->callback, obj, NULL, NULL);

			if(ayyi_client_remove_responder(AYYI_OBJECT_AUDIO_PART, AYYI_SET, obj.idx, data->handler)){
				g_free(data);
			}else{
				gwarn("cannot remove handler!");
			}
		}
	}
#endif

	g_return_if_fail(part);

	int64_t mu = ayyi_pos2mu(_pos);

	char buff[32]; ayyi_pos2bbst(_pos, buff); dbg (0, "split_pos=%s mu=%Lu=0x%016Lx %isamples", buff, mu, mu, ayyi_pos2samples(_pos));

	char start[32];
	char len[32];
	pos2bbst(&part->start, start);
	pos2bbst0(&part->length, len);
	dbg(0, "part.start=%s part.length=%s", start, len);

	AyyiAction* a = part_action_new(part, user_callback, user_data, "region split. idx=%u", part->ayyi->shm_idx);
	a->callback   = am_part__split_done;
#ifdef XAAA
	a->app_data   = ayyi_handler_data_new(user_callback, user_data);

#if 0
	struct _data* data = g_new0(struct _data, 1);
	data->callback = user_callback;
	data->handler = am_part__split_cb2;
	AyyiIdent id = {AYYI_OBJECT_AUDIO_PART, part->ayyi->shm_idx};
	ayyi_client_add_responder(id, AYYI_SET, am_part__split_handler, data);
#endif
#endif

	ayyi_set_obj(a, AYYI_OBJECT_PART, part->ayyi->shm_idx, AYYI_SPLIT, mu);
}


void
am_part__rename(AMPart* part, const char* new_name, AyyiHandler callback, gpointer user_data)
{
	g_return_if_fail(new_name);
	ASSERT_SONG_SHM;
	PF;

	void
	am_part__rename_done(AyyiAction* action)
	{
		PF;
		PartCallbackData* d = action->app_data;
		AMPart* part = callback_part;
		g_return_if_fail(part);
		if(!am_part__is_deleted(part)){
			part_label_sync(part);
			g_signal_emit_by_name (song->parts, "item-changed", part, CHANGE_NAME, NULL);
			call(d->user_callback, (AyyiIdent){action->obj.type, action->ret.idx}, &action->ret.error, d->user_data);
		}
		dbg (0, "new name=%s.", part->name);
		log_print(0, "part rename");
		am_part__unref(part);
	}

	AyyiAction* a = part_action_new(part, callback, user_data, "part rename. '%s'", new_name);
	a->callback = am_part__rename_done;

	ayyi_set_string(a, AYYI_OBJECT_PART, part->ayyi->shm_idx, new_name);
}


void
am_part__mute_toggle_async(AMPart* part)
{
	dbg (0, "(not yet implemented in songcore)");

	if(!ayyi.got_shm) return;

	void
	part_mute_toggle_cb(AyyiAction* action)
	{
		dbg (0, "FIXME pobj_id=%i", callback_part->ayyi->shm_idx);

		//AMPart* part = callback_part;
		if(action->ret.error){
			log_print(0, "error. mute failed.");
			return;
		}
	  
		//gnome_canvas_item_set(part->box, "muted", TRUE, NULL);
		log_print(LOG_OK, "part muted ok.");

		am_part__unref(callback_part);
	}

	uint32_t state = am_part__is_muted(part);

	AyyiAction* a = part_action_new(part, NULL, NULL, "part mute (%s)", part->name);
	a->callback   = part_mute_toggle_cb;

	ayyi_set_obj(a, part->ident.type, part->ayyi->shm_idx, AYYI_MUTE, state);
}


void
am_part__set_length(AMPart* part)
{
	g_return_if_fail(part);
	PF;

	void
	am_part__set_length_done(AyyiAction* action)
	{
		PF;

		AMPart* part = callback_part;
		g_return_if_fail(part);

		#if 0 //should be done in independent handler.
		g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_LEN, NULL);
		#endif

		am_part__unref(part);
	}

	GPos max; am_part__max_length(part, &max);
	pos_min(&part->length, &max);

	uint64_t mu = pos2mu(&part->length);

	char bbst[32]; pos2bbst0(&part->length, bbst);

	if(mu > PART_MAX_MU){ gerr ("part too long: %s (%Lu).", bbst, mu); return; }

	AyyiAction* a   = part_action_new(part, NULL, NULL, "part set length (%s %Limu)", bbst, mu);
	a->callback     = am_part__set_length_done;
	a->obj.type     = part->ident.type;
	a->obj_idx.idx1 = part->ayyi->shm_idx;

	dbg (0, "pod_index=%x", part->ayyi->shm_idx);

	ayyi_set_length(a, mu);
}


void
am_part__set_level_async(AMPart* part, float level)
{
	void
	part_set_level_cb(AyyiAction* action)
	{
		g_return_if_fail(action);
	}

	AyyiAction* a  = ayyi_action_new("part_set_level (%.2f)", level);
	a->callback    = part_set_level_cb;

	ayyi_set_float(a, AYYI_OBJECT_PART, part->ayyi->shm_idx, AYYI_PB_LEVEL, level);
}


void
am_part__set_fadein_async(AMPart* part, int fade_time)
{
	void
	part_set_fadein_cb(AyyiAction* action)
	{
		g_signal_emit_by_name (am_parts, "item-changed", callback_part, CHANGE_FADES, NULL);
		am_part__unref(callback_part);
	}

	AyyiAction* a = part_action_new(part, NULL, NULL, "part_set_fadein (%i)", fade_time);
	a->callback   = part_set_fadein_cb;

	part_set_fadein_cb(a); //temp!
}


void
am_part__set_fadeout_async(AMPart* part, int fade_time)
{
	void
	part_set_fadeout_cb(AyyiAction* action)
	{
		g_signal_emit_by_name (am_parts, "item-changed", callback_part, CHANGE_FADES, NULL);
		am_part__unref(callback_part);
	}

	AyyiAction* a = part_action_new(part, NULL, NULL, "part_set_fadeout (%i)", fade_time);
	a->callback   = part_set_fadeout_cb;

	part_set_fadeout_cb(a); //temp!
}


void
am_part__max_length(AMPart* part, GPos* max)
{
	//return the max allowable length for the given part.

	if(!PART_IS_AUDIO(part)){ pos_set(max, 4*400, 0, 0); return; }

	uint32_t source_length = part->pool_item->samplecount;
	samples2pos(source_length - part->region_start, max);
}


void
am_part__end_pos(AMPart* part, GPos* pos)
{
	*pos = part->start;
	if(!gpos_is_valid(&part->start)) gerr ("gpart start invalid");
	//char bbst1[64]; pos2bbst(pos, bbst1); dbg(0, "start=%s", bbst1);
	if(!gpos_is_valid(&part->length)) gerr ("gpart length invalid");
	//char bbst2[64]; pos2bbst0(&length, bbst2); dbg(0, "length=%s", bbst2);
	pos_add(pos, &part->length);
	//char bbst3[64]; pos2bbst(pos, bbst3); dbg(0, "end=%s", bbst3);
}


void
am_part__get_name(const AMPart* part, char* name)
{
	g_return_if_fail(part);

	g_strlcpy(name, part->name, AYYI_NAME_MAX);
}


static void
part_label_sync(AMPart* part)
{
	_am_part_sync_name(part);

	g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_NAME, NULL);
}


void
am_part__nudge_left(const AMPart* part, gboolean snap, QMap* q_map)
{
	GPos nudge, start;
	pos_copy(&start, &part->start);
	if(snap){
		q_to_prev(&start, q_map);
	}else{
		pos_set(&nudge, 0, 1, 0);
		pos_subtract(&start, &nudge);
	}
	am_part__move((AMPart*)part, &start, part->track, NULL, NULL);
}


void
am_part__nudge_right(const AMPart* part, gboolean snap, QMap* q_map)
{
	GPos pos, start;
	pos_copy(&start, &part->start);
	if(snap){
		q_to_next(&start, q_map);

		char bbst1[64]; pos2bbst((GPos*)&part->start, bbst1);
		char bbst2[64]; pos2bbst(&start, bbst2);
		dbg(0, "snap=%i %s-->%s", snap, bbst1, bbst2);
	}else{
		pos_set(&pos, 0, 1, 0);
		pos_add(&start, &pos);
	}
	am_part__move((AMPart*)part, &start, part->track, NULL, NULL);
}


char*
am_part__get_length_bbst(AMPart* part, char* bbst)
{
	g_return_val_if_fail(part, NULL);

	AyyiRegionBase* shared;
	if((shared = part_get_shared(part))){
		ayyi_samples2bbst(shared->length, bbst);
	}
	return bbst;
}


void
am_part__set_colour(AMPart* part, unsigned cidx, AyyiHandler handler, gpointer user_data)
{
	//should be usable by dnd (and others?)
	//-FIXME dont use for muting -part_redraw() should indicate muted status without the colour property changing.
	//arg 'cidx' is the colour index.  

	void part_set_colour_done(AyyiAction* action)
	{
		PF;

		AMPart* part = callback_part;
		g_return_if_fail(part);

		part->bg_colour = GPOINTER_TO_INT(((PartCallbackData*)action->app_data)->track);
		if(!part->fg_is_set){
			//the value set here is a cache to save the gui having to calculate it each time.
			part->fg_colour = am_part__get_auto_fg_colour(part);
		}

		g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_COLOUR, NULL);

		am_part__unref(part);
	}

	g_return_if_fail(part);

	if(cidx >= MAX_COLOUR){
		gwarn ("colouridx out of range: %i", cidx);
		cidx = 0;
	}

	AyyiAction* a  = part_action_new(part, handler, user_data, "part_colour_set (%i)", cidx);
	a->callback    = part_set_colour_done;
	((PartCallbackData*)a->app_data)->track = GINT_TO_POINTER(cidx);

	//ardour doesnt support region colour.
	part_set_colour_done(a); //FIXME we need to pipe this through the messaging system else action will not be freed.
}


void
am_part__get_fg_colour(AMPart* part, char* rgb, Palette* palette)
{
	//if the colour is not explicitly set, we use an automatically determined contrasting colour.

	//TODO change to take into account part->fg_is_set

	g_return_if_fail(part);

	if(part->fg_colour){
		sprintf(rgb, "#%06x", part->fg_colour >> 8);
	} else {
		uint32_t fg = am_colour_get_contrasting(palette, part->bg_colour) >> 8;
		sprintf(rgb, "#%06x", fg);
		dbg(3, "fg colour not set. contrasting=#%06x", fg);
	}
	dbg(3, "fg=%s", rgb);
}


static uint32_t
am_part__get_auto_fg_colour(AMPart* part)
{
	//if the colour is not explicitly set, we use an automatically determined contrasting colour.

	g_return_val_if_fail(part, 0);

	uint32_t fg = /*part->fg_colour
		? part->fg_colour
		: */am_colour_get_contrasting(&song->palette, part->bg_colour);

	return fg;
}


void
am_part__set_fg_colour(AMPart* part, uint32_t colour)
{
	part->fg_colour = colour;
	part->fg_is_set = true;

	g_signal_emit_by_name (am_parts, "item-changed", part, CHANGE_COLOUR, NULL);
}


gboolean
am_part__get_start_pos(AMPart* gpart, GPos* pos)
{
	g_return_val_if_fail(pos, false);

	AyyiRegionBase* region = part_get_shared(gpart);
	g_return_val_if_fail(region, false);

	samples2pos(region->position/* + region->start*/, pos);

	return TRUE;
}


#if 0
struct _filesource_shared*
ardour_region_get_filesource(uint32_t region_idx)
{
  printf("%s(): FIXME\n", __func__);

  //struct _region_shared* region = region_pod(gpart->ayyi->shm_idx);
  //ASSERT_POINTER_NULL(region, "region");

  //uint64_t source_id = region->source0;
  return NULL;
}


static void
part_get_engine_colour(AMPart* gpart, GdkColor* colour)
{
  ASSERT_POINTER(gpart, "gpart");
  gerr("FIXME");
}
#endif


float
am_part__get_level(AMPart* gpart)
{
	g_return_val_if_fail(gpart, 0.0);

	AyyiRegion* region = ayyi_song__audio_region_at(gpart->ayyi->shm_idx);
	g_return_val_if_fail(region, 0.0);

	return region->level;
}


uint32_t
am_part__get_fade_in(AMPart* part)
{
	g_return_val_if_fail(part, 0);
	if(!PART_IS_AUDIO(part)) return 0;

	AyyiRegion* region = (AyyiRegion*)part->ayyi;
	g_return_val_if_fail(region, 0);

	return region->fade_in_length;
}


uint32_t
am_part__get_fade_out(AMPart* part)
{
	g_return_val_if_fail(part, 0);
	if(!PART_IS_AUDIO(part)) return 0;

	AyyiRegion* region = (AyyiRegion*)part->ayyi;
	g_return_val_if_fail(region, 0);

	return region->fade_out_length;
}


gboolean
am_part__is_muted(AMPart* part)
{
	g_return_val_if_fail(part, false);
	PF;
	dbg(0, "FIXME add flags to midi_region");
	return 0;
	/*
	struct _AyyiRegionBase* region = PART_IS_AUDIO ? (struct _AyyiRegionBase*)region_pod(gpart->ayyi->shm_idx) : (struct _AyyiRegionBase*)midi_region_pod(gpart->ayyi->shm_idx);
	ASSERT_POINTER_FALSE(region, "region");
	return region->flags & 0x1;
	*/
}


bool
am_part__is_deleted(AMPart* part)
{
	g_return_val_if_fail(part, false);
	AyyiRegionBase* region = part_get_shared(part);
	g_return_val_if_fail(region, false);
	return region->flags & deleted;
}


const AMPart*
am_part__is_between(const AMPart* part, const AMPart* start, const AMPart* end)
{
	//if the start of Part is inbetween Start and End, it is returned, otherwise End is returned.
	//-if End is NULL, Part is returned if it is later than Start.
	//-if Start is NULL, Part is returned if it is early than End.

	//-used to find the next part after Start.
	//-fn is not useful, and is being replaced by part_get_previous() etc.

	g_return_val_if_fail(part, NULL);

	GPos start_pos = {0, 0, 0};
	if(start) start_pos = start->start;

	if(!pos_is_after(&part->start, &start_pos)) return end; // Part is too early.

	if(!end) return part;

	if(pos_is_before(&part->start, &end->start)){
		//printf("%s(): part=%i yes!\n", __func__, part->ayyi->shm_idx);
		return part; // yes, Part is inbetween start and end.
	}
	else{
		//printf("%s(): part=%i end=%i - too late\n", __func__, part->ayyi->shm_idx, end->idx);
		return end;                                          // Part is too late.
	}
}


static AMTrack*
_am_part_get_ayyi_track(AMPart* part)
{
	g_return_val_if_fail(part->ayyi, NULL);

	AyyiRegionBase* region = part_get_shared(part);
	AMTrack* trk = am_song__get_track_for_part(region);
	return trk;
}


bool
am_part__sync(AMPart* gpart, AyyiPropType prop)
{
	// completely syncs the local part to the remote object.

	bool err = FALSE;

	g_return_val_if_fail(gpart->ident.type && (gpart->ident.type == AYYI_OBJECT_AUDIO_PART || gpart->ident.type == AYYI_OBJECT_MIDI_PART), false);
	AyyiRegionBase* part = part_get_shared(gpart);
	if(!part){ gerr ("cannot get ayyi part. idx=%i", gpart->ident.idx); return !(err = TRUE); }
	if(part->flags & deleted){ dbg(0, "deleted! idx=%i", gpart->ident.idx); return false; }
	g_return_val_if_fail(PART_IS_AUDIO(gpart) ? ayyi_region__index_ok(part->shm_idx) : ayyi_song__midi_region_index_ok(part->shm_idx), false);
	if(!(part->id)){ gerr ("illegal id 0. idx=%i", part->shm_idx); return false; }

	_am_part_sync_name(gpart);
	_am_part_sync_pos(gpart);

	if(PART_IS_AUDIO(gpart)){
		//inset:
		AyyiRegion* audio_region = (AyyiRegion*)part;
		gpart->region_start = audio_region->start; //hmmm - true? yes i think so.
		dbg (2, "name='%s' id=%Lu position=%u region_start=%u", gpart->name, part->id, part->position, audio_region->start);
		if(gpart->region_start && gpart->pool_item){
			if(audio_region->start + audio_region->length > gpart->pool_item->samplecount){
				gerr ("bad inset! region_start=%u, len=%u n_samples=%Li", audio_region->start, audio_region->length, gpart->pool_item->samplecount);
				gpart->region_start = 0;
				err = TRUE;
			}
		}

		gpart->pool_item = am_part__get_pool_item(gpart);
		// let the caller mark the part bad if neccesary
		if(!gpart->pool_item && gpart->status != PART_PENDING){ gwarn("part has no pool item. idx=%i", part->shm_idx); /*part->bad = TRUE;*/ return false; }
	}

	am_part__sync_length(gpart);

	if(prop == AYYI_NO_PROP || prop & AYYI_TRACK){
		AyyiPlaylist* playlist = ayyi_song__playlist_at(part->playlist);
		if(playlist){
			gpart->track = am_song__get_track_by_playlist(playlist);
			if(!gpart->track){
				gerr ("cannot get track for part! part.idx=%i playlist=%s", gpart->ayyi->shm_idx, playlist->name);
				err = true;
			}
		} else {
			gerr ("part has no playlist! idx=%i", gpart->ayyi->shm_idx);
			err = TRUE;
		}
	}

	return !err;
}


AMPart*
am_part__init_from_index(AMPart* part, AyyiIdx shm_index)
{
	/*
	initialise data for a part which is new to the gui but already exists in core.

	*/

	part->ayyi = (AyyiRegionBase*)ayyi_song__audio_region_at(shm_index);
	g_return_val_if_fail(part->ayyi, NULL);
	part->ident.idx = shm_index;
	part->ident.type = AYYI_OBJECT_AUDIO_PART;

	if(!am_part__sync(part, AYYI_NO_PROP)){ gerr ("init failed - aborting..."); return NULL; }

	if(part->length.beat > MAX_PART_LENGTH_BEATS){
		gwarn ("part length too long (%i).", part->length.beat);
		pos_set(&part->length, 100, 0, 0);
	}

	am_part_check_bounds(part);

	dbg (2, "'%s' shm_index=%u start=.", part->name, shm_index);
	return part;
}


static gboolean
_am_part_sync_name(AMPart* gpart)
{
	g_return_val_if_fail(gpart, false);

	AyyiRegionBase* region = part_get_shared(gpart);

	snprintf(gpart->name, 63, "%s", region->name);

	return true;
}


static bool
_am_part_sync_pos(AMPart* part)
{
	//update the AMPart object with the current server value for the part start-time

	g_return_val_if_fail(part, false);

	AyyiAudioRegion* region = (AyyiAudioRegion*)(part_get_shared(part));
	g_return_val_if_fail(region, false);

	jack_nframes_t samples = region->position;
	dbg (2, "part->start=%u (position=%u start(offset)=%u)", samples, region->position, region->start);
	samples2pos(samples, &part->start);

	return true;
}


static gboolean
am_part_check_bounds(AMPart* gpart)
{
	g_return_val_if_fail(gpart, false);
	if(!PART_IS_AUDIO(gpart)) return FALSE;
	g_return_val_if_fail(gpart->pool_item, false);
	if(!PART_IS_AUDIO(gpart)) return TRUE;

	AyyiRegion* shared = ayyi_song__audio_region_at(gpart->ayyi->shm_idx);
	if(shared){
		uint32_t source_length = gpart->pool_item->samplecount;
		if(source_length && (shared->length > source_length)){ gwarn ("region is too long for the source! %i '%s', region_length=%u source_length=%u", shared->shm_idx, shared->name, shared->length, source_length); return FALSE; }

		return TRUE;
	}

	return FALSE;
}


void
am_part__sync_length(AMPart* gpart)
{
	g_return_if_fail(gpart);

	AyyiRegionBase* region = part_get_shared(gpart);
	g_return_if_fail(region);

	//ensure that because of the loss of precision here, we dont set zero length
	GPos min = {0, 0, 1};
	samples2pos(MAX(pos2samples(&min), region->length), &gpart->length);
}


static PoolItem*
am_part__get_pool_item(AMPart* part)
{
	g_return_val_if_fail(part, NULL);

	AyyiAudioRegion* region = (AyyiAudioRegion*)part->ayyi;//ayyi_song__audio_region_at(part->ayyi->shm_idx);
	g_return_val_if_fail(region, NULL);

	uint64_t source_id = region->source0;

	return am_pool__get_item_by_id(source_id);
}


void
am_part__set_note_selection(AMPart* gpart, GList* selection)
{
	dbg(2, "note=%i", selection ? ((MidiNote*)selection->data)->note : -1);

	MidiPart* part = (MidiPart*)gpart;
	if(part->note_selection) g_list_free(part->note_selection);
	part->note_selection = selection;
}


bool
am_part__is_audio(void* _part, gpointer user_data)
{
	AMPart* part = _part;
	return PART_IS_AUDIO(part);
}


bool
am_part__is_at_position(AMPart* part, gpointer _pos)
{
	AyyiSongPos* pos = _pos;
	GPos s;
	songpos_ayyi2gui(&s, pos);
	//char bbst1[64], bbst2[64]; pos2bbst(&s, bbst1); pos2bbst(&part->start, bbst2); dbg(0, "pos=%s - %s", bbst1, bbst2);
	if(!gpos_cmp(&s, &part->start)){ //TODO probably should use the ayyi part start instead
		//dbg(0, "same");
		return true;
	}
	return false;
}


bool
am_part__has_pool_item(AMPart* part, gpointer _pool_item)
{
	PoolItem* pool_item = _pool_item;
	return part->pool_item == pool_item;
}


bool
am_part__has_waveform(AMPart* part, gpointer waveform)
{
	return part->pool_item->waveform == (Waveform*)waveform;
}


bool
am_part__has_track(AMPart* part, AMTrack* track)
{
	g_return_val_if_fail(part, false);

	return part->track == track;
}


bool
am_part__has_id(AMPart* part, uint64_t* id)
{
	g_return_val_if_fail(part && id, false);

	return part->ayyi->id == *id;
}


void
am_part__on_palette_change (AMPart* part)
{
	if(!part->fg_is_set){
		part->fg_colour = am_part__get_auto_fg_colour(part);
	}
}


void
am_part__print(AMPart* part)
{
	g_return_if_fail(part);

	char bbst1[64], bbst2[64];
	pos2bbst(&part->start, bbst1);
	pos2bbst(&part->length, bbst2);
	uint64_t id = 0;
	char* source = part->pool_item ? part->pool_item->leafname : "";
	AyyiRegionBase* region = part->ayyi;
	if(region){
		id = region->id;
	}
	printf("%s(): %02i start=%s length=%s id=%Lu pool_item=%s %s%s%s\n", __func__, part->ident.idx, bbst1, bbst2, id, source, bold, part->name, white); 
}


const char*
am_part__print_status(AMPartStatus status)
{
	static char unk[32];
  
	switch (status){
#define CASE(x) case PART_##x: return "PART_"#x
		CASE (OK);
		CASE (LOCAL);
		CASE (PENDING);
		CASE (BAD);
#undef CASE
		default:
			snprintf (unk, 31, "UNKNOWN PART STATUS (%d)", status);
			return unk;
	}
	return NULL;
}


